/*
 *  tree.c
 *  VisBoardGame
 *
 *  Created by fredericVernier on 2/4/07.
 *  Copyright 2007 Frederic Vernier. All rights reserved.
 *
 */

#include "tree.h"


#include <stdlib.h>
#include <stddef.h>
#include <assert.h>
#include "tree.h"


/*
 * Initialize a tree object supplied by the client such that it becomes a valid
 * empty tree. If the tree is to be ``unbounded'', the maxcount should be
 * specified as treecount_t_MAX, or, alternately, as -1. The value zero
 * is not permitted.
 */

tree_t *tree_init(tree_t *tree, treecount_t maxcount) {
  assert (maxcount != 0);
  tree->root = NULL;
  tree->nodecount = 0;
  tree->maxcount = maxcount;
  return tree;
}

/*
 * Dynamically allocate a tree object using malloc(), and initialize it so that
 * it is a valid empty tree. If the tree is to be ``unbounded'', the maxcount
 * should be specified as treecount_t_MAX, or, alternately, as -1.
 */

tree_t *tree_create(treecount_t maxcount) {
  tree_t *newt = malloc(sizeof *newt);
  if (newt) {
    assert (maxcount != 0);
    newt->root = NULL;
    newt->nodecount = 0;
    newt->maxcount = maxcount;
  }
  return newt;
}

/*
 * Destroy a dynamically allocated tree object.
 * The client must remove the nodes first.
 */

void tree_destroy(tree_t *tree) {
  assert (tree_isempty(tree));
  free(tree);
}

/*
 * Free all of the nodes of a tree. The tree must contain only
 * dynamically allocated nodes. After this call, the tree
 * is empty.
 */
void tree_destroy_nodes_r(tnode_t *node) {
  if (node->children==NULL) return;

  lnode_t *ln = list_first(node->children);
  while (ln !=  &node->children->list_nilnode) {
    tnode_t *tn = (tnode_t *)ln->list_data;
    tree_destroy_nodes_r(tn);
    ln = ln->list_next;
  }
  list_destroy_nodes(node->children);
  list_destroy(node->children);
}
void tree_destroy_nodes(tree_t *tree) {
  tree_destroy_nodes_r(tree->root);
  tree_init(tree, tree->maxcount);
}


/*
 * Insert the node ``new'' into the tree as a child of ``this'' node.
 */
void tree_ins_under(tree_t *tree, tnode_t *new, tnode_t *this) {

  assert (new != NULL);
  assert (!tree_contains(tree, new));
  assert (!tnode_is_in_a_tree(new));
  assert (tree->nodecount + 1 > tree->nodecount);

  list_append(this->children, lnode_create(new));
  tree->nodecount++;

  assert (tree->nodecount <= tree->maxcount);
}


/*
 * Insert the node ``new'' into the tree as a brother of ``this'' node.
 */

void tree_ins_after(tree_t *tree, tnode_t *new, tnode_t *this) {
  assert (new != NULL);
  assert (!tree_contains(tree, new));
  assert (tree_contains(tree, this));
  assert (!tnode_is_in_a_tree(new));
  assert (tnode_is_in_a_tree(this));
  assert (tree->root!=this);
  assert (tree->nodecount + 1 > tree->nodecount);

  list_append(this->father->children, lnode_create(new));
  tree->nodecount++;

  assert (tree->nodecount <= tree->maxcount);
}


/**
 * comparizon function
 */
int compare(const void *a, const void *b) {
  return (a != b);
}// compare() --------------------------------------------------------------------------------


/*
 * Delete the given node/leaf from the tree.
 */
tnode_t *tree_delete(tree_t *tree, tnode_t *del) {
  assert (tree_contains(tree, del));
  assert (del->children==NULL || list_count(del->children)==0);

  lnode_t *ln = list_find(del->father->children, del, &compare);

  list_delete(del->father->children, ln);
  tree->nodecount--;
  del->father = NULL;
  if (del->children!=NULL)
    list_destroy(del->children);

  return del;
}


/*
 * For each node in the tree, execute the given function. The tree,
 * current node and the given context pointer are passed on each
 * call to the function.
 */

void tree_process(tree_t *tree, void *context, void (* function)(tree_t *tree, tnode_t *tnode, void *context)) {
  // to be done
  assert(0==1);
}


/*
 * Dynamically allocate a tree node and assign it the given piece of data.
 */
tnode_t *tnode_create(void *data) {
  tnode_t *new = malloc(sizeof *new);
  if (new) {
    new->data = data;
    new->father = NULL;
    new->children = NULL;
  }
  return new;
}


/*
 * Initialize a user-supplied tnode.
 */
tnode_t *tnode_init(tnode_t *tnode, void *data) {
  tnode->data = data;
  tnode->father = NULL;
  tnode->children = NULL;
  return tnode;
}


/*
 * Destroy a dynamically allocated node.
 */
void tnode_destroy(tnode_t *tnode){
  assert (!tnode_is_in_a_tree(tnode));
  free(tnode);
}


/*
 * Determine whether the given tree contains the given node.
 */
 int tree_contains_r(tnode_t *node, tnode_t *snode) {
  if (node->children==NULL) return 0;

  int found = 0;
  lnode_t *ln = list_first(node->children);
  while (!found && ln !=  &node->children->list_nilnode) {
    tnode_t *tn = (tnode_t *)ln->list_data;
    if (tn==snode)
      found=1;
    else
      found = tree_contains_r(tn, snode);
    ln = ln->list_next;
  }
  return found;
}
int tree_contains(tree_t *tree, tnode_t *node) {
  return tree_contains_r(tree->root, node);
}


/*
 * A more generalized variant of tree_transfer. This one removes a
 * ``slice'' from the source tree and appends it to the destination
 * tree.
 */
void tree_extract(tree_t *dest, tree_t *source, tnode_t *first, tnode_t *last) {
  // to be done
  assert (0==1);
}


/*
 * Split off a trailing sequence of nodes from the source tree and relocate
 * them to the tail of the destination tree. The trailing sequence begins
 * with node ``first'' and terminates with the last node of the source
 * tree. The nodes are added to the end of the new tree in their original
 * order.
 */
void tree_transfer(tree_t *dest, tree_t *source, tnode_t *first) {
  // to be done
  assert (0==1);
}

void tree_merge(tree_t *dest, tree_t *sour,
  int compare (const void *, const void *)) {
  // to be done
  assert (0==1);
}

void tree_sort(tree_t *tree, int compare(const void *, const void *)) {
  // to be done
  assert (0==1);
}

tnode_t *tree_find(tree_t *tree, const void *key, int compare(const void *, const void *)) {
  // to be done
  assert (0==1);
}


/*
 * Return 1 if the tree is in sorted order, 0 otherwise
 */

int tree_is_sorted(tree_t *tree, int compare(const void *, const void *)) {
  // to be done
  assert (0==1);
}


/*
 * Return 1 if the tree is empty, 0 otherwise
 */
int tree_isempty(tree_t *tree) {
  return tree->nodecount == 0;
}


/*
 * Return 1 if the tree is full, 0 otherwise
 * Permitted only on bounded trees.
 */
int tree_isfull(tree_t *tree) {
  return tree->nodecount == tree->maxcount;
}


/*
 * Add the given node at the end of the tree
 */
void tree_append(tree_t *tree, tnode_t *node) {
  // to be done
  assert (0==1);
}


/*
 * Add the given node at the beginning of the tree.
 */
void tree_prepend(tree_t *tree, tnode_t *node) {
  // to be done
  assert (0==1);
}


/*
 * Retrieve the first node of the tree
 */
tnode_t *tree_root(tree_t *tree) {
  return tree->root;
}


/*
 * Retrieve the count of nodes in the tree
 */
treecount_t tree_count(tree_t *tree) {
  return tree->nodecount;
}


/*
 * Associate a data item with the given node.
 */
void tnode_put(tnode_t *tnode, void *data) {
  tnode->data = data;
}


/*
 * Retrieve the data item associated with the node.
 */
void *tnode_get(tnode_t *tnode) {
  return tnode->data;
}


/*
 * Return 1 if the tnode is in some tree, otherwise return 0.
 */

int tnode_is_in_a_tree(tnode_t *tnode) {
  return (tnode->father != NULL);
}


int tree_verify_r(tnode_t *node) {
  if (node->children==NULL) return 1;

  int ok = 1;
  lnode_t *ln = list_first(node->children);
  while (ok && ln !=  &node->children->list_nilnode) {
    tnode_t *tn = (tnode_t *)ln->list_data;
    if (tn->father!=node)
      ok = 0;
    else
      ok = tree_verify_r(tn);
    ln = ln->list_next;
  }
  return ok;
}
int tree_verify(tree_t *tree) {
  treecount_t count = tree_count(tree);

  if (count > tree->maxcount)
    return 0;

  if (tree->root!=NULL)
    return 0;

  if (tree_verify_r(tree->root))
    return 0;

  return 1;
}


